//
//  KBGCardGenerator.m
//  Keynote Bingo Generator
//
//  Created by Peter Hosey on 2007-01-01.
//  Copyright 2007 Peter Hosey. All rights reserved.
//

#import "KBGCardGenerator.h"

static void pageSetupSheetIsDone(PMPrintSession printSession, WindowRef documentWindow, Boolean accepted);
static PMSheetDoneUPP pageSetupSheetIsDoneUPP = NULL;

@interface KBGCardGenerator (CarbonPrintingManagerFakery)
- (void)pageSetupDidEnd:(PMPrintSession)session
			 returnCode:(int)returnCode
			contextInfo:(void *)contextInfo;
@end

static void messageCallbackForCGPSConverter(void *info, CFStringRef message);

static KBGCardGenerator *singleton = nil;

@implementation KBGCardGenerator

+ (void)initialize {
	if(!pageSetupSheetIsDoneUPP) {
		pageSetupSheetIsDoneUPP = NewPMSheetDoneUPP(pageSetupSheetIsDone);
	}

	if(!singleton) {
		singleton = [[KBGCardGenerator alloc] init];
	}
}

+ (KBGCardGenerator *)sharedGenerator {
	return singleton;
}

- init {
	if((self = [super init])) {
		OSStatus err;

		err = PMCreateSession(&printSession);
		if(err == noErr) {
			err = PMCreatePageFormat(&pageFormat);
			if(err == noErr) {
				err = PMSessionDefaultPageFormat(printSession, pageFormat);

				useSeedNumber = NO;
				incrementSeedNumber = YES;
				numberOfCardsPerPage = 1U;

				//I expect that the average user will only generate one card. If not, it shouldn't be a *huge* penalty, especially since Foundation is free to ignore my suggestions anyway.
				tempFiles = [[NSMutableSet alloc] initWithCapacity:1U];
				[[NSNotificationCenter defaultCenter]
					addObserver:self
					   selector:@selector(applicationWillTerminate:)
						   name:NSApplicationWillTerminateNotification
						 object:NSApp];

				[NSBundle loadNibNamed:@"KBGCardGenerator" owner:self];
			}
		}
	}
	return self;
}
- (void)dealloc {
	[super dealloc];
}

- (void)awakeFromNib {
	if(![window setFrameUsingName:[window frameAutosaveName]]) {
		[window center];
	}

	WindowRef windowRef = [window windowRef];

	//Set the IgnoreClicks attribute because if we don't, whatever overlay window Cocoa is returning from -windowRef may intercept further clicks on the Page Setup. This is an intermittent — but highly annoying — bug.
	ChangeWindowAttributes(windowRef, kWindowIgnoreClicksAttribute, 0U);

	//Run the Page Setup sheet on our window.
	PMSessionUseSheets(printSession, windowRef, pageSetupSheetIsDoneUPP);
}

#pragma mark Accessors

- (unsigned)seedNumber {
	return seedNumber;
}
- (void)setSeedNumber:(unsigned)newSeedNumber {
	seedNumber = newSeedNumber;
}

- (BOOL)useSeedNumber {
	return useSeedNumber;
}
- (void)setUseSeedNumber:(BOOL)flag {
	useSeedNumber = flag;
}

- (BOOL)incrementSeedNumber {
	return incrementSeedNumber;
}
- (void)setIncrementSeedNumber:(BOOL)flag {
	incrementSeedNumber = flag;
}

#pragma mark Actions

- (IBAction)useSeedOrNotChanged:sender {
	[self setUseSeedNumber:[[sender selectedCell] tag]];
}
- (IBAction)furtherSeedsMethodChanged:sender {
	[self setIncrementSeedNumber:[[sender selectedCell] tag]];
}
- (IBAction)nupChanged:sender {
	numberOfCardsPerPage = [[sender selectedCell] tag];
}

- (IBAction)runPageLayout:sender {
	//Disable the button to prevent the user from multiple-clicking and getting us both a sheet and one or more modal dialogs.
	[pageSetupButton setEnabled:NO];

	Boolean accepted_unused;
	PMSessionPageSetupDialog(printSession, pageFormat, &accepted_unused);

	//Testing revealed that PMSessionPageSetupDialog does not return until the sheet has been fully unrolled. Handy.
	[pageSetupButton setEnabled:YES];
}
- (void)pageSetupDidEnd:(PMPrintSession)session
			 returnCode:(int)returnCode
			contextInfo:(void *)contextInfo
{
	[window performSelector:@selector(makeKeyAndOrderFront:)
				 withObject:nil
				 afterDelay:0.00000001];
}

- (IBAction)bingoMePlease:sender {
	//Get orientation and paper size from printSession/pageFormat.
	PMOrientation orientation;
	PMGetOrientation(pageFormat, &orientation);

	PMRect paperRectFromCarbonPrintingManager;
	PMGetAdjustedPaperRect(pageFormat, &paperRectFromCarbonPrintingManager);
	NSRect paperRect = {
		.origin = { paperRectFromCarbonPrintingManager.bottom, paperRectFromCarbonPrintingManager.left },
		.size   = {
			.width  = paperRectFromCarbonPrintingManager.right  - paperRectFromCarbonPrintingManager.left,
			.height = paperRectFromCarbonPrintingManager.bottom - paperRectFromCarbonPrintingManager.top
		}
	};
	double scale;
	PMGetScale(pageFormat, &scale);
	scale /= 100.0; //Convert from percentage to fraction.

	NSString *templatePath = [[NSBundle mainBundle] pathForResource:@"bingocardtemplate" ofType:@"ps"];
	NSError *error = nil;
	NSMutableString *template = [NSMutableString stringWithContentsOfFile:templatePath
																 encoding:NSUTF8StringEncoding
																	error:&error];
	if(error) {
		[window presentError:error
			  modalForWindow:window
					delegate:nil
		  didPresentSelector:NULL
				 contextInfo:NULL];
	} else {
		BOOL isLandscape, isFlipped;
		switch(orientation) {
		case kPMPortrait:
		default:
			isLandscape = NO;
			isFlipped   = NO;
			break;

		case kPMReversePortrait:
			isLandscape = NO;
			isFlipped   = YES;
			break;

		case kPMLandscape:
			isLandscape = YES;
			isFlipped   = NO;
			break;

		case kPMReverseLandscape:
			isLandscape = YES;
			isFlipped   = YES;
			break;
		}

		if(isLandscape) {
			float temp = paperRect.size.width;
			paperRect.size.width  = paperRect.size.height;
			paperRect.size.height = temp;
		}

		[template replaceOccurrencesOfString:@"%%WIDTH%%"
								  withString:[NSString stringWithFormat:@"%u", (unsigned)paperRect.size.width]
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];
		[template replaceOccurrencesOfString:@"%%HEIGHT%%"
								  withString:[NSString stringWithFormat:@"%u", (unsigned)paperRect.size.height]
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];

		NSMutableString *transformCode = [NSMutableString stringWithFormat:
			@"%E %E translate %% Crop to paper size\n"
			@"%E dup scale %% Scale to %.2f%% as directed by Page Setup\n",
			(paperRect.size.width  - 612.0f) * 0.5f, //612pt = 8.5″
			(paperRect.size.height - 792.0f) * 0.5f, //792pt = 11″
			scale,
			scale * 100.0f];
		if(isLandscape) {
			[transformCode appendFormat:
				@"%% Landscape orientation.\n"
				@"%E 0 translate\n"
				@"-90 rotate\n",
				paperRect.size.width];
		}
		if(isFlipped) {
			[transformCode appendFormat:
				@"%% Flipped orientation.\n"
				@"0 %E translate\n"
				@"1 -1 scale\n",
				paperRect.size.height];
		}

		[template replaceOccurrencesOfString:@"%%CTM%%\n"
								  withString:transformCode
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];

		union {
			UnsignedWide uw;
			unsigned long long ull;
		} timeSinceStartup;
		Microseconds(&timeSinceStartup.uw);
		unsigned long millisecondsSinceStartup = ((timeSinceStartup.ull / 1000ULL) % ULONG_MAX);

		[template replaceOccurrencesOfString:@"%%SEED%%\n"
								  withString:(useSeedNumber
											  ? [NSString stringWithFormat:@"/seednumber %u def\n", seedNumber]
											  : [NSString stringWithFormat:@"/seednumber %lu def %% User provided no explicit seed; this is the time in milliseconds since the user started up. It is more reliable than PostScript's realtime operator, which is not guaranteed to be highly variant or indeed even positive. Negative numbers just don't look good after \"Seed #\" on the page.\n", millisecondsSinceStartup])
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];
		[template replaceOccurrencesOfString:@"%%INCREMENTSEED%%"
								  withString:((useSeedNumber && incrementSeedNumber) //Don't localize these; they're PostScript literals.
											  ? @"true"
											  : @"false")
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];

		[template replaceOccurrencesOfString:@"%%NUP%%"
								  withString:((numberOfCardsPerPage > 1U)
											  ? [NSString stringWithFormat:@"-%uup", numberOfCardsPerPage]
											  : @"")
									 options:NSLiteralSearch
									   range:(NSRange){ 0U, [template length] }];

		NSString *tmpFolder = nil;
		FSRef ref;
		OSStatus err = FSFindFolder(kLocalDomain, kChewableItemsFolderType, /*createFolder*/ true, &ref);
		if(err == noErr) {
			NSURL *tmpFolderURL = [(NSURL *)CFURLCreateFromFSRef(kCFAllocatorDefault, &ref) autorelease];
			tmpFolder = [tmpFolderURL path];
		}
		if(!tmpFolder)
			tmpFolder = NSTemporaryDirectory();
		[[template dataUsingEncoding:NSISOLatin1StringEncoding] writeToFile:[tmpFolder stringByAppendingPathComponent:@"bingocard.eps"] atomically:NO];

		/*Example filenames:
		 *	Implicit seed, 1-up: bingocard-1234567890.pdf
		 *	Explicit seed, 4-up: bingocard-4up-1.pdf
		 */
		NSString *filename = [NSString stringWithFormat:@"bingocard%@-%lu.pdf",
			//Include “-NUMup” (for NUM=numberOfCardsPerPage) in the filename when numberOfCardsPerPage > 1.
			((numberOfCardsPerPage > 1U) ? [NSString stringWithFormat:@"-%uup", numberOfCardsPerPage] : @""),
			//Insert our initial seed — whatever it is — into the filename.
			useSeedNumber ? (unsigned long)seedNumber : millisecondsSinceStartup];
		NSString *savePath = [tmpFolder stringByAppendingPathComponent:filename];
		[tempFiles addObject:savePath];
		NSURL *saveURL = [NSURL fileURLWithPath:savePath];

		NSMutableArray *errorMessageLines = [NSMutableArray array];
		static const struct CGPSConverterCallbacks callbacks = {
			.version = 0U,
			.noteMessage = messageCallbackForCGPSConverter
			//We don't need any of the other callbacks. According to C99 (§6.7.8 paragraph 10), they will be initialized to NULL for us.
		};
		CGPSConverterRef converter = CGPSConverterCreate(/*info*/ errorMessageLines,
														 &callbacks,
														 /*options*/ NULL);
		CGDataProviderRef provider = CGDataProviderCreateWithCFData((CFDataRef)[template dataUsingEncoding:NSISOLatin1StringEncoding]);
		CGDataConsumerRef consumer = CGDataConsumerCreateWithURL((CFURLRef)saveURL);
		bool success = CGPSConverterConvert(converter,
											provider,
											consumer,
											/*options*/ NULL);
		CFRelease(consumer);
		CFRelease(provider);
		CFRelease(converter);

		if(!success) {
			NSAlert *alert = [NSAlert alertWithMessageText:NSLocalizedString(@"Creation of the PDF document from the PostScript source failed.", /*comment*/ @"Message text for the alert presented when CGPSConverter fails.")
											 defaultButton:nil
										   alternateButton:nil
											   otherButton:nil
								 informativeTextWithFormat:@"%@",
				[errorMessageLines componentsJoinedByString:@"\n"]];
			[alert beginSheetModalForWindow:window
							  modalDelegate:nil
							 didEndSelector:NULL
								contextInfo:NULL];
		} else {
			NSDocumentController *docController = [NSDocumentController sharedDocumentController];

			NSString *type = [docController typeForContentsOfURL:saveURL
														   error:&error];
			if(error) goto presentError;

			NSDocument *doc = [docController makeDocumentForURL:saveURL
											  withContentsOfURL:saveURL
														 ofType:type
														  error:&error];
			if(error) goto presentError;

			[docController addDocument:doc];
			[doc makeWindowControllers];
			[doc updateChangeCount:NSChangeDone];
			[doc showWindows];

			if(error) {
			presentError:
				[window presentError:error
					  modalForWindow:window
							delegate:nil
				  didPresentSelector:NULL
						 contextInfo:NULL];
			}
		}
	}
}

#pragma mark Notifications

- (void)applicationWillTerminate:(NSNotification *)notification {
	//Clean up after ourselves.
	NSEnumerator *tempFilesEnum = [tempFiles objectEnumerator];
	NSString *tempFilePath;
	while((tempFilePath = [tempFilesEnum nextObject])) {
		NSFileManager *mgr = [NSFileManager defaultManager];
		[mgr removeFileAtPath:tempFilePath handler:nil];
	}
}

@end

static void pageSetupSheetIsDone(PMPrintSession printSession, WindowRef documentWindow, Boolean accepted) {
	[singleton pageSetupDidEnd:printSession
					returnCode:accepted ? NSOKButton : NSCancelButton
				   contextInfo:documentWindow];
}

static void messageCallbackForCGPSConverter(void *info, CFStringRef message) {
	//info is our array of lines, created in -bingoMePlease:.
	[(NSMutableArray *)info addObject:(NSString *)message];
}
